package com.btk5h.skriptdb;

import static java.sql.Types.*;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.Instant;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.IntStream;

import ch.njol.skript.util.Date;
import ch.njol.skript.util.Time;

public class SkriptValueWrapper {
	
	private static final Map<Class<?>, SkriptValueWrapper> wrappersBySkript = new HashMap<>();
	private static final Map<Integer, SkriptValueWrapper> wrappersBySql = new HashMap<>();
	
	
	public static final SkriptValueWrapper FALLBACK = register(Object.class, Function.identity(), ResultSet::getObject);
	public static final SkriptValueWrapper TEXT_VALUE = register(String.class, SkriptValueWrapper::asString, SkriptValueWrapper::getString, CHAR, VARCHAR, LONGVARCHAR, CLOB);
	public static final SkriptValueWrapper BOOLEAN_VALUE = register(Boolean.class, SkriptValueWrapper::asBoolean, ResultSet::getBoolean, BOOLEAN);
	public static final SkriptValueWrapper LONG_VALUE = register(Long.class, x -> x, SkriptValueWrapper::getLong, BIT, TINYINT, SMALLINT, INTEGER, BIGINT);
	public static final SkriptValueWrapper DOUBLE_VALUE = register(Double.class, x -> x, SkriptValueWrapper::getDouble, FLOAT, REAL, DOUBLE, NUMERIC, DECIMAL);
	public static final SkriptValueWrapper DATE_VALUE = register(String.class, SkriptValueWrapper::noParser, SkriptValueWrapper::getString, DATE);
	public static final SkriptValueWrapper TIME_VALUE = register(Time.class, SkriptValueWrapper::asTime, SkriptValueWrapper::getTime, TIME);
	public static final SkriptValueWrapper TIMESTAMP_VALUE = register(Date.class, SkriptValueWrapper::asTimestamp, SkriptValueWrapper::getTimestamp, TIMESTAMP);
	public static final SkriptValueWrapper BINARY_VALUE = register(String.class, SkriptValueWrapper::noParser, SkriptValueWrapper::getBinary, TIMESTAMP);
	
	
	
	private final Class<?> skriptType;
	private final Function<Object, Object> skToSqlMapper;
	private final ResultFetcher<Object> sqlToSkMapper;
	private final int[] validSqlTypes;
	
	
	@SuppressWarnings("unchecked")
	private <T> SkriptValueWrapper(Class<T> skriptType, Function<T, Object> skToSql, ResultFetcher<? super T> sqlToSk, int... validSqlTypes) {
		this.skriptType = skriptType;
		this.skToSqlMapper = (Function<Object, Object>) skToSql;
		this.sqlToSkMapper = (ResultFetcher<Object>) sqlToSk;
		this.validSqlTypes = IntStream.of(validSqlTypes)
				.sorted()
				.toArray();
	}
	
	
	public Object asSql(Object obj) {
		return this.skToSqlMapper.apply(obj);
	}
	
	
	public boolean isSkApplicable(Class<?> type) {
		return this.skriptType.isAssignableFrom(type);
	}
	
	
	public boolean isSqlApplicable(int sqlType) {
		return Arrays.binarySearch(this.validSqlTypes, sqlType) >= 0;
	}
	
	
	public Object asSkript(ResultSet rs, int column) throws SQLException {
		return this.sqlToSkMapper.fetch(rs, column);
	}
	
	
	public static SkriptValueWrapper getBySQLType(int type) {
		return wrappersBySql.getOrDefault(type, FALLBACK);
	}
	
	
	public static SkriptValueWrapper getBySkriptType(Class<?> type) {
		return wrappersBySkript.getOrDefault(type, FALLBACK);
	}
	
	
	public static SkriptValueWrapper getBySQLLType(int type) {
		return switch (type) {
			//case CHAR, VARCHAR, LONGVARCHAR, CLOB -> null; // string
			//case BOOLEAN -> null; // boolean
			//case BIT, TINYINT, SMALLINT, INTEGER, BIGINT -> null; // long
			//case FLOAT, REAL, DOUBLE, NUMERIC, DECIMAL -> null; // double
			//case DATE -> null; // string
			//case TIME -> null; // time
			//case TIMESTAMP -> null; // Date
			//case BINARY, VARBINARY, LONGVARBINARY, BLOB -> null; // hex String
			//case NULL, OTHER, JAVA_OBJECT, DISTINCT, STRUCT, ARRAY -> null; // just whatever it gives to me
			default -> null;
		};
	}
	
	
	private static <T> SkriptValueWrapper register(Class<T> skriptType, Function<T, Object> skToSql, ResultFetcher<? super T> sqlToSk, int... validSqlTypes) {
		SkriptValueWrapper wrapper = new SkriptValueWrapper(skriptType, skToSql, sqlToSk, validSqlTypes);
		wrappersBySkript.putIfAbsent(skriptType, wrapper);
		for (int i : validSqlTypes) {
			wrappersBySql.putIfAbsent(i, wrapper);
		}
		return wrapper;
	}
	
	
	private static String getString(ResultSet rs, int column) throws SQLException {
		return rs.getString(column);
	}
	
	private static String asString(Object obj) {
		return obj != null ? String.valueOf(obj) : null;
	}
	
	
	private static Long getLong(ResultSet rs, int column) throws SQLException {
		var val = rs.getLong(column);
		if (!rs.wasNull()) {
			return val;
		}
		return null;
	}
	
	
	private static Double getDouble(ResultSet rs, int column) throws SQLException {
		var val = rs.getDouble(column);
		if (!rs.wasNull()) {
			return val;
		}
		return null;
	}
	
	
	private static final int TICKS_PER_HOUR = 1000;
	private static final double TICKS_PER_MINUTE = 1000. / 60;
	private static final int HOUR_ZERO = 6 * TICKS_PER_HOUR;
	
	private static Time getTime(ResultSet rs, int column) throws SQLException {
		var time = rs.getTime(column);
		if (time != null) {
			var local = time.toLocalTime();
			var minutes = local.getMinute();
			var hours = local.getHour();
			return new Time((int) Math.round(
					hours * TICKS_PER_HOUR
					- HOUR_ZERO
					+ minutes * TICKS_PER_MINUTE));
		}
		return null;
	}
	
	private static Object asTime(Time time) {
		if (time != null) {
			return new java.sql.Time(time.getTime());
		}
		return null;
	}
	
	
	private static Date getTimestamp(ResultSet rs, int column) throws SQLException {
		return new Date(rs.getTimestamp(column).getTime());
	}
	
	private static Object asTimestamp(Date date) {
		return Timestamp.from(Instant.ofEpochMilli(date.getTimestamp()));
	}
	
	private static Object asBoolean(Object obj) {
		if (obj instanceof Number n) {
			return n.doubleValue() > 0;
		} else if (obj instanceof Boolean b) {
			return b;
		}
		return null;
	}
	
	
	private static String getBinary(ResultSet rs, int column) throws SQLException {
		byte[] bytes = rs.getBytes(column);
		if (!rs.wasNull()) {
			StringBuilder sb = new StringBuilder(bytes.length);
			for (int i = 0; i < bytes.length; i++) {
				sb.append(String.format("%02X", bytes[i]));
			}
			return sb.toString();
		}
		return null;
	}
	
	
	private static <T, U> U noParser(T t) {
		throw new IllegalStateException("This value shouldn't have a parser");
	}
	
	
	private interface ResultFetcher<T> {
		public T fetch(ResultSet rs, int index) throws SQLException;
	}

}
